/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include "mlx5dv_dr.h"
#include "mlx5_ifc.h"

/* Each matcher has a priority this is the number of allowed
 * priorities per flow table
 *
 * TODO: Support unlimited number of priorities - Should be done after
 * E-Switch merge
 */
#define DR_TBL_MAX_PRIORITY		128
#define DR_TBL_ROOT_MAX_PRIORITY	16

static void dr_tbl_uninit_nic(struct dr_tbl_rx_tx *nic_tbl)
{
	dr_htbl_put(nic_tbl->s_anchor);
}

static void dr_tbl_uninit_fdb(struct mlx5dv_dr_tbl *tbl)
{
	dr_tbl_uninit_nic(&tbl->rx);
	dr_tbl_uninit_nic(&tbl->tx);
}

static void dr_tbl_uninit(struct mlx5dv_dr_tbl *tbl)
{
	switch (tbl->ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		dr_tbl_uninit_nic(&tbl->rx);
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		dr_tbl_uninit_nic(&tbl->tx);
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		dr_tbl_uninit_fdb(tbl);
		break;
	default:
		errno = EINVAL;
		return;
	}

	free(tbl->prio);
}

static int dr_tbl_init_nic(struct mlx5dv_dr_ns *ns,
			   struct dr_tbl_rx_tx *nic_tbl)
{
	struct dr_ns_rx_tx *nic_ns = nic_tbl->nic_ns;
	struct dr_htbl_connect_info info;
	int ret;

	nic_tbl->s_anchor = dr_ste_htbl_alloc(ns->ste_icm_pool,
					      DR_CHUNK_SIZE_1,
					      DR_STE_LU_TYPE_DONT_CARE,
					      0);
	if (!nic_tbl->s_anchor)
		return errno;

	info.type = CONNECT_MISS;
	info.miss_icm_addr = nic_ns->default_icm_addr;
	ret = dr_ste_htbl_init_and_postsend(ns, nic_ns, nic_tbl->s_anchor,
					    &info, true);
	if (ret) {
		errno = EINVAL;
		goto free_s_anchor;
	}

	dr_htbl_get(nic_tbl->s_anchor);

	return 0;

free_s_anchor:
	dr_ste_htbl_free(nic_tbl->s_anchor);
	return errno;
}

static int dr_tbl_init_fdb(struct mlx5dv_dr_tbl *tbl)
{
	int ret;

	ret = dr_tbl_init_nic(tbl->ns, &tbl->rx);
	if (ret)
		return ret;

	ret = dr_tbl_init_nic(tbl->ns, &tbl->tx);
	if (ret)
		goto destroy_rx;

	return 0;

destroy_rx:
	dr_tbl_uninit_nic(&tbl->rx);
	return ret;
}

static int dr_tbl_init(struct mlx5dv_dr_tbl *tbl)
{
	int ret = 0;

	list_head_init(&tbl->matcher_list);
	tbl->prio = calloc(tbl->max_prio, sizeof(struct mlx5dv_dr_matcher *));
	if (!tbl->prio)
		return ENOMEM;

	switch (tbl->ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		tbl->table_type = FS_FT_NIC_RX;
		tbl->rx.nic_ns = &tbl->ns->info.rx;
		ret = dr_tbl_init_nic(tbl->ns, &tbl->rx);
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		tbl->table_type = FS_FT_NIC_TX;
		tbl->tx.nic_ns = &tbl->ns->info.tx;
		ret = dr_tbl_init_nic(tbl->ns, &tbl->tx);
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		tbl->table_type = FS_FT_FDB;
		tbl->rx.nic_ns = &tbl->ns->info.rx;
		tbl->tx.nic_ns = &tbl->ns->info.tx;
		ret = dr_tbl_init_fdb(tbl);
		break;
	default:
		errno = EINVAL;
		ret = errno;
		break;
	}

	if (ret)
		goto free_prio;

	return 0;

free_prio:
	free(tbl->prio);
	return ret;
}
static int dr_tbl_destory_devx_tbl(struct mlx5dv_dr_tbl *tbl)
{
	return dr_devx_destroy_flow_table(tbl->devx_obj);
}

static int dr_tbl_create_devx_tbl(struct mlx5dv_dr_tbl *tbl)
{
	uint64_t icm_addr_rx = 0;
	uint64_t icm_addr_tx = 0;

	if (tbl->rx.s_anchor)
		icm_addr_rx = tbl->rx.s_anchor->chunk->icm_addr;

	if (tbl->tx.s_anchor)
		icm_addr_tx = tbl->tx.s_anchor->chunk->icm_addr;

	tbl->devx_obj = dr_devx_create_flow_table(tbl->ns->ctx,
						  tbl->table_type,
						  icm_addr_rx,
						  icm_addr_tx);
	if (!tbl->devx_obj)
		return EINVAL;

	return 0;
}

struct mlx5dv_dr_tbl *mlx5dv_dr_create_ft(struct mlx5dv_dr_ns *ns,
					  uint32_t level)
{
	struct mlx5dv_dr_tbl *tbl;
	int ret;

	if (level && !ns->info.supp_sw_steering) {
		errno = EOPNOTSUPP;
		return NULL;
	}

	tbl = calloc(1, sizeof(*tbl));
	if (!tbl)
		return NULL;

	tbl->ns = ns;
	tbl->level = level;
	tbl->max_prio = level ? DR_TBL_MAX_PRIORITY : DR_TBL_ROOT_MAX_PRIORITY;
	pthread_mutex_init(&tbl->mutex, NULL);
	list_head_init(&tbl->matcher_list);
	atomic_init(&tbl->refcount, 1);

	if (!dr_is_root_table(tbl)) {
		ret = dr_tbl_init(tbl);
		if (ret)
			goto free_tbl;

		ret = dr_tbl_create_devx_tbl(tbl);
		if (ret)
			goto uninit_tbl;
	}

	atomic_fetch_add(&ns->refcount, 1);
	return tbl;

uninit_tbl:
	dr_tbl_uninit(tbl);
free_tbl:
	free(tbl);
	return NULL;
}

int mlx5dv_dr_destroy_ft(struct mlx5dv_dr_tbl *tbl)
{
	if (atomic_load(&tbl->refcount) > 1)
		return EBUSY;

	if (!dr_is_root_table(tbl)) {
		dr_tbl_destory_devx_tbl(tbl);
		dr_tbl_uninit(tbl);
	}

	atomic_fetch_sub(&tbl->ns->refcount, 1);
	pthread_mutex_destroy(&tbl->mutex);
	free(tbl);

	return 0;
}
